"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from . import annotations


class BindingKey(object):
    """The key for a binding."""

    def __init__(self, name, annotation):
        """Initializer.

        Args:
          name: the name of the bound arg
          annotation: an Annotation
        """
        self._name = name
        self._annotation = annotation

    def __repr__(self):
        return '<{0}>'.format(self)

    def __str__(self):
        return 'the binding name "{0}" ({1})'.format(
            self._name, self.annotation_as_adjective())

    def annotation_as_adjective(self):
        return self._annotation.as_adjective()

    def __eq__(self, other):
        return (isinstance(other, BindingKey) and
                self._name == other._name and
                self._annotation == other._annotation)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self._name) ^ hash(self._annotation)


def new(arg_name, annotated_with=None):
    """Creates a BindingKey.

    Args:
      arg_name: the name of the bound arg
      annotation: an Annotation, or None to create an unannotated binding key
    Returns:
      a new BindingKey
    """
    if annotated_with is not None:
        annotation = annotations.Annotation(annotated_with)
    else:
        annotation = annotations.NO_ANNOTATION
    return BindingKey(arg_name, annotation)



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import six
import inspect

from . import errors

# To be removed once this fix is included in six
# https://github.com/benjaminp/six/issues/155
try:
    import collections.abc as collections_abc
except ImportError:  # python <3.3
    import collections as collections_abc


def items(dict_instance):
    return six.iteritems(dict_instance)


def is_sequence(arg_value):
    return isinstance(arg_value, collections_abc.Sequence)


def is_string(arg_value):
    return isinstance(arg_value, six.string_types)


def is_constructor_defined(cls):
    if six.PY3:
        return inspect.isfunction(cls.__init__)
    return inspect.ismethod(cls.__init__)


def get_method_args(fn):
    if six.PY3:
        spec = inspect.getfullargspec(fn)
        return spec.args, spec.varargs, spec.varkw, spec.defaults
    arg_names, varargs, keywords, defaults = inspect.getargspec(fn)
    return arg_names, varargs, keywords, defaults


def verify_callable(fn, arg_name):
    if not callable(fn):
        raise errors.WrongArgTypeError(arg_name, 'callable', type(fn).__name__)


def verify_subclasses(seq, required_superclass, arg_name):
    if not isinstance(seq, collections_abc.Sequence):
        raise errors.WrongArgTypeError(
            arg_name,
            'sequence (of subclasses of {0})'.format(
                required_superclass.__name__),
            type(seq).__name__)
    for idx, elt in enumerate(seq):
        if not isinstance(elt, required_superclass):
            raise errors.WrongArgElementTypeError(
                arg_name, idx,
                'subclass of {0}'.format(required_superclass.__name__),
                type(elt).__name__)


def verify_module_types(modules, arg_name):
    _verify_types(inspect.ismodule, modules, arg_name, 'module')


def verify_class_types(seq, arg_name):
    _verify_types(inspect.isclass, seq, arg_name, 'class')


def verify_class_type(elt, arg_name):
    _verify_type(inspect.isclass, elt, arg_name, 'class')


def _assert_sequence(seq, arg_name, type_name):
    if not is_sequence(seq):
        raise errors.WrongArgTypeError(
            arg_name, 'sequence (of {0})'.format(type_name), type(seq).__name__)


def _verify_types(fn_checker, seq, arg_name, type_name):
    _assert_sequence(seq, arg_name, type_name)
    for idx, elt in enumerate(seq):
        if not fn_checker(elt):
            raise errors.WrongArgElementTypeError(
                arg_name, idx, type_name, type(elt).__name__)


def _verify_type(fn_checker, elt, arg_name, type_name):
    if not fn_checker(elt):
        raise errors.WrongArgTypeError(
            arg_name, type_name, type(elt).__name__)



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import decorator

from . import errors
from . import support


def copy_args_to_internal_fields(fn):
    """Copies the initializer args to internal member fields.

    This is a decorator that applies to __init__.
    """
    return _copy_args_to_fields(fn, 'copy_args_to_internal_fields', '_')


def copy_args_to_public_fields(fn):
    """Copies the initializer args to public member fields.

    This is a decorator that applies to __init__.
    """
    return _copy_args_to_fields(fn, 'copy_args_to_public_fields', '')


def _copy_args_to_fields(fn, decorator_name, field_prefix):
    if fn.__name__ != '__init__':
        raise errors.DecoratorAppliedToNonInitError(
            decorator_name, fn)
    arg_names, varargs, unused_keywords, unused_defaults = (
        support.get_method_args(fn))
    if varargs is not None:
        raise errors.PargsDisallowedWhenCopyingArgsError(
            decorator_name, fn, varargs)

    def CopyThenCall(fn_to_wrap, self, *pargs, **kwargs):
        for index, parg in enumerate(pargs, start=1):
            setattr(self, field_prefix + arg_names[index], parg)
        for kwarg, kwvalue in support.items(kwargs):
            setattr(self, field_prefix + kwarg, kwvalue)
        fn_to_wrap(self, *pargs, **kwargs)

    return decorator.decorator(CopyThenCall, fn)



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


class Annotation(object):
    """A binding annotation."""

    def __init__(self, annotation_obj):
        """Initializer.

        Args:
          annotation_obj: the annotation object, which can be any object that
              implements __eq__() and __hash__()
        """
        self._annotation_obj = annotation_obj

    def as_adjective(self):
        """Returns the annotation as an adjective phrase.

        For example, if the annotation object is '3', then the annotation
        adjective phrase is 'annotated with "3"'.

        Returns:
          an annotation adjective phrase
        """
        return 'annotated with "{0}"'.format(self._annotation_obj)

    def __repr__(self):
        return '<{0}>'.format(self.as_adjective())

    def __eq__(self, other):
        return (isinstance(other, Annotation) and
                self._annotation_obj == other._annotation_obj)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self._annotation_obj)


class _NoAnnotation(object):
    """A polymorph for Annotation but that actually means "no annotation"."""

    def as_adjective(self):
        return 'unannotated'

    def __repr__(self):
        return '<{0}>'.format(self.as_adjective())

    def __eq__(self, other):
        return isinstance(other, _NoAnnotation)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return 0


NO_ANNOTATION = _NoAnnotation()



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from . import locations


class Error(Exception):
    pass


class AmbiguousArgNameError(Error):

    def __init__(self, injection_site_desc, binding_key, bindings):
        Error.__init__(
            self, 'when injecting {0}, {1} ambiguously refers to any'
            ' of:\n{2}'.format(
                injection_site_desc, binding_key, '\n'.join(
                    '  {0}'.format(b.get_binding_target_desc_fn())
                    for b in bindings)))


class BadDependencyScopeError(Error):

    def __init__(self, injection_site_desc,
                 from_scope_id, to_scope_id, binding_key):
        Error.__init__(
            self, 'when injecting {0} in {1}, scope {2} is not usable from'
            ' scope {3}'.format(
                binding_key, injection_site_desc, to_scope_id, from_scope_id))


class ConfigureMethodMissingArgsError(Error):

    def __init__(self, configure_fn, possible_args):
        Error.__init__(
            self, 'binding spec method {0} must have at least'
            ' one of the expected args {1}'.format(
                locations.get_name_and_loc(configure_fn), possible_args))


class ConflictingExplicitBindingsError(Error):

    def __init__(self, colliding_bindings):
        Error.__init__(
            self, 'multiple explicit bindings for same binding name:\n'
            '{0}'.format('\n'.join('  {0}'.format(b)
                                   for b in colliding_bindings)))


class ConflictingRequiredBindingError(Error):

    def __init__(self, required_binding, colliding_bindings):
        Error.__init__(
            self, 'conflicting implicit bindings for binding required at {0}'
            ' for {1}:\n{2}'.format(
                required_binding.require_loc, required_binding.binding_key,
                '\n'.join('  {0}'.format(b) for b in colliding_bindings)))


class CyclicInjectionError(Error):

    def __init__(self, binding_stack):
        Error.__init__(
            self, 'cyclic injections:\n{0}'.format(
                '\n'.join('  {0}'.format(b) for b in binding_stack)))


class DecoratorAppliedToNonInitError(Error):

    def __init__(self, decorator_name, fn):
        Error.__init__(
            self, '@{0} cannot be applied to non-initializer {1}'.format(
                decorator_name, locations.get_name_and_loc(fn)))


class DirectlyPassingInjectedArgsError(Error):

    def __init__(self, duplicated_args, injection_site_desc, provider_fn):
        Error.__init__(
            self, 'somewhere in {0}, injected args {1} passed directly when'
            ' calling {2}'.format(
                injection_site_desc, list(duplicated_args),
                locations.get_name_and_loc(provider_fn)))


class DuplicateDecoratorError(Error):

    def __init__(self, decorator_name, second_decorator_loc):
        Error.__init__(
            self, 'at {0}, @{1} cannot be applied twice'.format(
                second_decorator_loc, decorator_name))


class EmptyBindingSpecError(Error):

    def __init__(self, binding_spec):
        Error.__init__(
            self, 'binding spec {0} at {1} must have either a configure()'
            ' method or a provider method but has neither'.format(
                binding_spec.__class__.__name__,
                locations.get_loc(binding_spec.__class__)))


class EmptyProvidesDecoratorError(Error):

    def __init__(self, at_provides_loc):
        Error.__init__(
            self, '@provides() at {0} needs at least one non-default'
            ' arg'.format(at_provides_loc))


class EmptySequenceArgError(Error):

    def __init__(self, call_site_loc, arg_name):
        Error.__init__(
            self, 'expected non-empty sequence arg {0} at {1}'.format(
                arg_name, call_site_loc))


class InjectingNoneDisallowedError(Error):

    def __init__(self, proviser_desc):
        Error.__init__(
            self, 'cannot inject None (returned from {0}) because'
            ' allow_injecting_none=False'.format(proviser_desc))


class InvalidBindingTargetError(Error):

    def __init__(self, binding_loc, binding_key, binding_target,
                 expected_type_str):
        Error.__init__(
            self, '{0} cannot be bound to {1} at {2} because the latter is of'
            ' type {3}, not {4}'.format(
                binding_key, binding_target, binding_loc,
                type(binding_target).__name__, expected_type_str))


class MissingRequiredBindingError(Error):

    def __init__(self, required_binding):
        Error.__init__(
            self, 'at {0}, binding required for {1}, but no such binding was'
            ' ever created'.format(required_binding.require_loc,
                                   required_binding.binding_key))


class MultipleAnnotationsForSameArgError(Error):

    def __init__(self, arg_binding_key, decorator_loc):
        Error.__init__(
            self, 'multiple annotations for {0} at {1}'.format(
                arg_binding_key, decorator_loc))


class MultipleBindingTargetArgsError(Error):

    def __init__(self, binding_loc, binding_key, arg_names):
        Error.__init__(
            self, 'multiple binding target args {0} given for {1} at'
            ' {2}'.format(arg_names, binding_key, binding_loc))


class NoBindingTargetArgsError(Error):

    def __init__(self, binding_loc, binding_key):
        Error.__init__(
            self, 'no binding target arg given for {0} at {1}'.format(
                binding_key, binding_loc))


class NoRemainingArgsToInjectError(Error):

    def __init__(self, decorator_loc):
        Error.__init__(
            self, 'at {0}, all args are declared passed directly and therefore'
            ' no args will be injected; call the method directly'
            ' instead?'.format(decorator_loc))


class NoSuchArgError(Error):

    def __init__(self, call_site_loc, arg_name):
        Error.__init__(self, 'at {0}, no such arg named {1}'.format(
            call_site_loc, arg_name))


# TODO(kurts): replace NoSuchArgToInjectError with NoSuchArgError.
class NoSuchArgToInjectError(Error):

    def __init__(self, decorator_loc, arg_binding_key, fn):
        Error.__init__(
            self, 'cannot inject {0} into {1} at {2}: no such arg name'.format(
                arg_binding_key, fn.__name__, decorator_loc))


class NonExplicitlyBoundClassError(Error):

    def __init__(self, provide_loc, cls):
        Error.__init__(
            self, 'at {0}, cannot instantiate class {1}, since it is not'
            ' explicitly marked as injectable and only_use_explicit_bindings'
            ' is set to True'.format(provide_loc, cls.__name__))


class NothingInjectableForArgError(Error):

    def __init__(self, binding_key, injection_site_desc):
        Error.__init__(
            self, 'when injecting {0}, nothing injectable for {1}'.format(
                injection_site_desc, binding_key))


class OnlyInstantiableViaProviderFunctionError(Error):

    def __init__(self, injection_site_fn, arg_binding_key, binding_target_desc):
        Error.__init__(
            self, 'when injecting {0}, {1} cannot be injected, because its'
            ' provider, {2}, needs at least one directly passed arg'.format(
                locations.get_name_and_loc(injection_site_fn),
                arg_binding_key, binding_target_desc))


class OverridingDefaultScopeError(Error):

    def __init__(self, scope_id):
        Error.__init__(
            self, 'cannot override default scope {0}'.format(scope_id))


class PargsDisallowedWhenCopyingArgsError(Error):

    def __init__(self, decorator_name, fn, pargs_arg_name):
        Error.__init__(
            self, 'decorator @{0} cannot be applied to {1} with *{2}'.format(
                decorator_name, locations.get_name_and_loc(fn), pargs_arg_name))


class TooManyArgsToInjectDecoratorError(Error):

    def __init__(self, decorator_loc):
        Error.__init__(
            self, 'at {0}, cannot specify both arg_names and'
            ' all_except'.format(decorator_loc))


class UnknownScopeError(Error):

    def __init__(self, scope_id, binding_loc):
        Error.__init__(self, 'unknown scope ID {0} in binding created at'
                       ' {1}'.format(scope_id, binding_loc))


class WrongArgElementTypeError(Error):

    def __init__(self, arg_name, idx, expected_type_desc, actual_type_desc):
        Error.__init__(
            self, 'wrong type for element {0} of arg {1}: expected {2} but got'
            ' {3}'.format(idx, arg_name, expected_type_desc, actual_type_desc))


class WrongArgTypeError(Error):

    def __init__(self, arg_name, expected_type_desc, actual_type_desc):
        Error.__init__(
            self, 'wrong type for arg {0}: expected {1} but got {2}'.format(
                arg_name, expected_type_desc, actual_type_desc))



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import re
import inspect
import threading

from . import binding_keys
from . import decorators
from . import errors
from . import locations
from . import providing
from . import scoping


class Binding(object):

    def __init__(self, binding_key, proviser_fn, get_binding_target_desc_fn,
                 scope_id, get_binding_loc_fn):
        self.binding_key = binding_key
        self.proviser_fn = proviser_fn
        self.get_binding_target_desc_fn = get_binding_target_desc_fn
        self.scope_id = scope_id
        self._get_binding_loc_fn = get_binding_loc_fn

    def __str__(self):
        return 'the binding at {0}, from {1} to {2}, in "{3}" scope'.format(
            self._get_binding_loc_fn(), self.binding_key,
            self.get_binding_target_desc_fn(), self.scope_id)


def _handle_explicit_binding_collision(
        colliding_binding, binding_key_to_binding, *pargs):
    other_binding = binding_key_to_binding[colliding_binding.binding_key]
    raise errors.ConflictingExplicitBindingsError(
        [colliding_binding, other_binding])


def _handle_implicit_binding_collision(
        colliding_binding, binding_key_to_binding,
        collided_binding_key_to_bindings):
    binding_key = colliding_binding.binding_key
    bindings = collided_binding_key_to_bindings.setdefault(
        binding_key, set())
    bindings.add(binding_key_to_binding[binding_key])
    del binding_key_to_binding[binding_key]


def _get_binding_key_to_binding_maps(bindings, handle_binding_collision_fn):
    binding_key_to_binding = {}
    collided_binding_key_to_bindings = {}
    for binding_ in bindings:
        binding_key = binding_.binding_key
        if binding_key in binding_key_to_binding:
            handle_binding_collision_fn(
                binding_, binding_key_to_binding,
                collided_binding_key_to_bindings)
        if binding_key in collided_binding_key_to_bindings:
            collided_binding_key_to_bindings[binding_key].add(binding_)
        else:
            binding_key_to_binding[binding_key] = binding_
    return binding_key_to_binding, collided_binding_key_to_bindings


def get_overall_binding_key_to_binding_maps(bindings_lists):

    """bindings_lists from lowest to highest priority.  Last item in
    bindings_lists is assumed explicit.

    """
    binding_key_to_binding = {}
    collided_binding_key_to_bindings = {}

    for index, bindings in enumerate(bindings_lists):
        is_final_index = (index == (len(bindings_lists) - 1))
        handle_binding_collision_fn = {
            True: _handle_explicit_binding_collision,
            False: _handle_implicit_binding_collision}[is_final_index]
        this_binding_key_to_binding, this_collided_binding_key_to_bindings = (
            _get_binding_key_to_binding_maps(
                bindings, handle_binding_collision_fn))
        for good_binding_key in this_binding_key_to_binding:
            collided_binding_key_to_bindings.pop(good_binding_key, None)
        binding_key_to_binding.update(this_binding_key_to_binding)
        collided_binding_key_to_bindings.update(
            this_collided_binding_key_to_bindings)

    return binding_key_to_binding, collided_binding_key_to_bindings


class BindingMapping(object):

    def __init__(self, binding_key_to_binding,
                 collided_binding_key_to_bindings):
        self._binding_key_to_binding = binding_key_to_binding
        self._collided_binding_key_to_bindings = (
            collided_binding_key_to_bindings)

    def verify_requirements(self, required_bindings):
        for required_binding in required_bindings:
            required_binding_key = required_binding.binding_key
            if required_binding_key not in self._binding_key_to_binding:
                if (required_binding_key in
                    self._collided_binding_key_to_bindings):
                    raise errors.ConflictingRequiredBindingError(
                        required_binding,
                        self._collided_binding_key_to_bindings[
                            required_binding_key])
                else:
                    raise errors.MissingRequiredBindingError(required_binding)

    def get(self, binding_key, injection_site_desc):
        if binding_key in self._binding_key_to_binding:
            return self._binding_key_to_binding[binding_key]
        elif binding_key in self._collided_binding_key_to_bindings:
            raise errors.AmbiguousArgNameError(
                injection_site_desc, binding_key,
                self._collided_binding_key_to_bindings[binding_key])
        else:
            raise errors.NothingInjectableForArgError(
                binding_key, injection_site_desc)


def default_get_arg_names_from_class_name(class_name):
    """Converts normal class names into normal arg names.

    Normal class names are assumed to be CamelCase with an optional leading
    underscore.  Normal arg names are assumed to be lower_with_underscores.

    Args:
      class_name: a class name, e.g., "FooBar" or "_FooBar"
    Returns:
      all likely corresponding arg names, e.g., ["foo_bar"]
    """
    parts = []
    rest = class_name
    if rest.startswith('_'):
        rest = rest[1:]
    while True:
        m = re.match(r'([A-Z][a-z]*|[0-9][a-z0-9]*)(.*)', rest)
        if m is None:
            break
        parts.append(m.group(1))
        rest = m.group(2)
    if not parts:
        return []
    return ['_'.join(part.lower() for part in parts)]


def get_explicit_class_bindings(
        classes,
        get_arg_names_from_class_name=default_get_arg_names_from_class_name):
    explicit_bindings = []
    for cls in classes:
        if decorators.is_explicitly_injectable(cls):
            for arg_name in get_arg_names_from_class_name(cls.__name__):
                explicit_bindings.append(new_binding_to_class(
                    binding_keys.new(arg_name), cls, scoping.DEFAULT_SCOPE,
                    lambda cls=cls: locations.get_loc(cls)))
    return explicit_bindings


def get_provider_bindings(
        binding_spec, known_scope_ids,
        get_arg_names_from_provider_fn_name=(
            providing.default_get_arg_names_from_provider_fn_name)):
    provider_bindings = []
    fns = inspect.getmembers(binding_spec, lambda x: inspect.ismethod(x))
    for _, fn in fns:
        default_arg_names = get_arg_names_from_provider_fn_name(fn.__name__)
        fn_bindings = get_provider_fn_bindings(fn, default_arg_names)
        for binding in fn_bindings:
            if binding.scope_id not in known_scope_ids:
                raise errors.UnknownScopeError(
                    binding.scope_id, locations.get_name_and_loc(fn))
        provider_bindings.extend(fn_bindings)
    return provider_bindings


def get_implicit_class_bindings(
        classes,
        get_arg_names_from_class_name=(
            default_get_arg_names_from_class_name)):
    implicit_bindings = []
    for cls in classes:
        arg_names = get_arg_names_from_class_name(cls.__name__)
        for arg_name in arg_names:
            implicit_bindings.append(new_binding_to_class(
                binding_keys.new(arg_name), cls, scoping.DEFAULT_SCOPE,
                lambda cls=cls: locations.get_loc(cls)))
    return implicit_bindings


class Binder(object):

    def __init__(self, collected_bindings, scope_ids):
        self._collected_bindings = collected_bindings
        self._scope_ids = scope_ids
        self._lock = threading.Lock()
        self._class_bindings_created = []

    def bind(self, arg_name, annotated_with=None,
             to_class=None, to_instance=None, in_scope=scoping.DEFAULT_SCOPE):
        if in_scope not in self._scope_ids:
            raise errors.UnknownScopeError(
                in_scope, locations.get_back_frame_loc())
        binding_key = binding_keys.new(arg_name, annotated_with)
        specified_to_params = [
            'to_class' if to_class is not None else None,
            'to_instance' if to_instance is not None else None]
        specified_to_params = [x for x in specified_to_params if x is not None]
        if not specified_to_params:
            binding_loc = locations.get_back_frame_loc()
            raise errors.NoBindingTargetArgsError(binding_loc, binding_key)
        elif len(specified_to_params) > 1:
            binding_loc = locations.get_back_frame_loc()
            raise errors.MultipleBindingTargetArgsError(
                binding_loc, binding_key, specified_to_params)

        # TODO(kurts): this is such a hack; isn't there a better way?
        if to_class is not None:
            @decorators.annotate_arg('_pinject_class', (to_class, in_scope))
            @decorators.provides(annotated_with=annotated_with,
                                 in_scope=in_scope)
            def provide_it(_pinject_class):
                return _pinject_class
            with self._lock:
                self._collected_bindings.extend(
                    get_provider_fn_bindings(provide_it, [arg_name]))
                if (to_class, in_scope) not in self._class_bindings_created:
                    back_frame_loc = locations.get_back_frame_loc()
                    self._collected_bindings.append(new_binding_to_class(
                        binding_keys.new('_pinject_class',
                                         (to_class, in_scope)),
                        to_class, in_scope, lambda: back_frame_loc))
                    self._class_bindings_created.append((to_class, in_scope))
        else:
            back_frame_loc = locations.get_back_frame_loc()
            with self._lock:
                self._collected_bindings.append(new_binding_to_instance(
                    binding_key, to_instance, in_scope,
                    lambda: back_frame_loc))


def new_binding_to_class(binding_key, to_class, in_scope, get_binding_loc_fn):
    if not inspect.isclass(to_class):
        raise errors.InvalidBindingTargetError(
            get_binding_loc_fn(), binding_key, to_class, 'class')
    def Proviser(injection_context, obj_provider, pargs, kwargs):
        return obj_provider.provide_class(
            to_class, injection_context, pargs, kwargs)
    def GetBindingTargetDesc():
        return 'the class {0}'.format(locations.get_name_and_loc(to_class))
    return Binding(binding_key, Proviser, GetBindingTargetDesc, in_scope,
                   get_binding_loc_fn)


def new_binding_to_instance(
        binding_key, to_instance, in_scope, get_binding_loc_fn):
    def Proviser(injection_context, obj_provider, pargs, kwargs):
        if pargs or kwargs:
            raise TypeError('instance provider takes no arguments'
                            ' ({0} given)'.format(len(pargs) + len(kwargs)))
        return to_instance
    def GetBindingTargetDesc():
        return 'the instance {0!r}'.format(to_instance)
    return Binding(binding_key, Proviser, GetBindingTargetDesc, in_scope,
                   get_binding_loc_fn)


class BindingSpec(object):

    def configure(self, bind):
        raise NotImplementedError()

    def dependencies(self):
        return []

    def __eq__(self, other):
        return type(self) == type(other)

    def __hash__(self):
        return hash(type(self))


def get_provider_fn_bindings(provider_fn, default_arg_names):
    provider_decorations = decorators.get_provider_fn_decorations(
        provider_fn, default_arg_names)
    def Proviser(injection_context, obj_provider, pargs, kwargs):
        return obj_provider.call_with_injection(
            provider_fn, injection_context, pargs, kwargs)
    def GetBindingTargetDescFn():
        return 'the provider method {0}'.format(
            locations.get_name_and_loc(provider_fn))
    return [
        Binding(binding_keys.new(provider_decoration.arg_name,
                                 provider_decoration.annotated_with),
                Proviser, GetBindingTargetDescFn,
                provider_decoration.in_scope_id,
                lambda p_fn=provider_fn: locations.get_loc(p_fn))
        for provider_decoration in provider_decorations]



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from . import bindings
from . import decorators
from . import errors
from . import finding
from . import injection_contexts
from . import locations
from . import object_providers
from . import providing
from . import required_bindings as required_bindings_lib
from . import scoping
from . import support


def new_object_graph(
        modules=finding.ALL_IMPORTED_MODULES, classes=None, binding_specs=None,
        only_use_explicit_bindings=False, allow_injecting_none=False,
        configure_method_name='configure',
        dependencies_method_name='dependencies',
        get_arg_names_from_class_name=(
            bindings.default_get_arg_names_from_class_name),
        get_arg_names_from_provider_fn_name=(
            providing.default_get_arg_names_from_provider_fn_name),
        id_to_scope=None, is_scope_usable_from_scope=lambda _1, _2: True,
        use_short_stack_traces=True):
    """Creates a new object graph.

    Args:
      modules: the modules in which to search for classes for which to create
          implicit bindings; if None, then no modules; by default, all
          modules imported at the time of calling this method
      classes: the classes for which to create implicit bindings; if None (the
          default), then no classes
      binding_specs: the BindingSpec subclasses to get bindings and provider
          methods from; if None (the default), then no binding specs
      only_use_explicit_bindings: whether to use only explicit bindings (i.e.,
          created by binding specs or @pinject.injectable, etc.)
      allow_injecting_none: whether to allow a provider method to provide None
      configure_method_name: the name of binding specs' configure method
      dependencies_method_name: the name of binding specs' dependencies method
      get_arg_names_from_class_name: a function mapping a class name to a
          sequence of the arg names to which those classes should be
          implicitly bound (if any)
      get_arg_names_from_provider_fn_name: a function mapping a provider
          method name to a sequence of the arg names for which that method is
          a provider (if any)
      id_to_scope: a map from scope ID to the concrete Scope implementation
          instance for that scope
      is_scope_usable_from_scope: a function taking two scope IDs and
          returning whether an object in the first scope can be injected into
          an object from the second scope; by default, injection is allowed
          from any scope into any other scope
      use_short_stack_traces: whether to shorten the stack traces for
          exceptions that Pinject raises, so that they don't contain the
          innards of Pinject
    Returns:
      an ObjectGraph
    Raises:
      Error: the object graph is not creatable as specified

    """
    try:
        if modules is not None and modules is not finding.ALL_IMPORTED_MODULES:
            support.verify_module_types(modules, 'modules')
        if classes is not None:
            support.verify_class_types(classes, 'classes')
        if binding_specs is not None:
            support.verify_subclasses(
                binding_specs, bindings.BindingSpec, 'binding_specs')
        if get_arg_names_from_class_name is not None:
            support.verify_callable(get_arg_names_from_class_name,
                                    'get_arg_names_from_class_name')
        if get_arg_names_from_provider_fn_name is not None:
            support.verify_callable(get_arg_names_from_provider_fn_name,
                                    'get_arg_names_from_provider_fn_name')
        if is_scope_usable_from_scope is not None:
            support.verify_callable(is_scope_usable_from_scope,
                                    'is_scope_usable_from_scope')
        injection_context_factory = injection_contexts.InjectionContextFactory(
            is_scope_usable_from_scope)
        id_to_scope = scoping.get_id_to_scope_with_defaults(id_to_scope)
        bindable_scopes = scoping.BindableScopes(id_to_scope)
        known_scope_ids = id_to_scope.keys()

        found_classes = finding.find_classes(modules, classes)
        if only_use_explicit_bindings:
            implicit_class_bindings = []
        else:
            implicit_class_bindings = bindings.get_implicit_class_bindings(
                found_classes, get_arg_names_from_class_name)
        explicit_bindings = bindings.get_explicit_class_bindings(
            found_classes, get_arg_names_from_class_name)
        binder = bindings.Binder(explicit_bindings, known_scope_ids)
        required_bindings = required_bindings_lib.RequiredBindings()
        if binding_specs is not None:
            binding_specs = list(binding_specs)
            processed_binding_specs = set()
            while binding_specs:
                binding_spec = binding_specs.pop()
                if binding_spec in processed_binding_specs:
                    continue
                processed_binding_specs.add(binding_spec)
                all_kwargs = {'bind': binder.bind,
                              'require': required_bindings.require}
                has_configure = hasattr(binding_spec, configure_method_name)
                if has_configure:
                    configure_method = getattr(binding_spec, configure_method_name)
                    configure_kwargs = _pare_to_present_args(
                        all_kwargs, configure_method)
                    if not configure_kwargs:
                        raise errors.ConfigureMethodMissingArgsError(
                            configure_method, all_kwargs.keys())
                    try:
                        configure_method(**configure_kwargs)
                    except NotImplementedError:
                        has_configure = False
                dependencies = None
                if hasattr(binding_spec, dependencies_method_name):
                    dependencies_method = (
                        getattr(binding_spec, dependencies_method_name))
                    dependencies = dependencies_method()
                    binding_specs.extend(dependencies)
                provider_bindings = bindings.get_provider_bindings(
                    binding_spec, known_scope_ids,
                    get_arg_names_from_provider_fn_name)
                explicit_bindings.extend(provider_bindings)
                if (not has_configure and
                    not dependencies and
                    not provider_bindings):
                    raise errors.EmptyBindingSpecError(binding_spec)
        binding_key_to_binding, collided_binding_key_to_bindings = (
            bindings.get_overall_binding_key_to_binding_maps(
                [implicit_class_bindings, explicit_bindings]))
        binding_mapping = bindings.BindingMapping(
            binding_key_to_binding, collided_binding_key_to_bindings)
        binding_mapping.verify_requirements(required_bindings.get())
    except errors.Error as e:
        if use_short_stack_traces:
            raise e
        else:
            raise

    is_injectable_fn = {True: decorators.is_explicitly_injectable,
                        False: (lambda cls: True)}[only_use_explicit_bindings]
    obj_provider = object_providers.ObjectProvider(
        binding_mapping, bindable_scopes, allow_injecting_none)
    return ObjectGraph(
        obj_provider, injection_context_factory, is_injectable_fn,
        use_short_stack_traces)


def _pare_to_present_args(kwargs, fn):
    arg_names, _, _, _ = support.get_method_args(fn)
    return {arg: value
            for arg, value in support.items(kwargs) if arg in arg_names}


class ObjectGraph(object):
    """A graph of objects instantiable with dependency injection."""

    def __init__(self, obj_provider, injection_context_factory,
                 is_injectable_fn, use_short_stack_traces):
        self._obj_provider = obj_provider
        self._injection_context_factory = injection_context_factory
        self._is_injectable_fn = is_injectable_fn
        self._use_short_stack_traces = use_short_stack_traces

    def provide(self, cls):
        """Provides an instance of the given class.

        Args:
          cls: a class (not an instance)
        Returns:
          an instance of cls
        Raises:
          Error: an instance of cls is not providable
        """
        support.verify_class_type(cls, 'cls')
        if not self._is_injectable_fn(cls):
            provide_loc = locations.get_back_frame_loc()
            raise errors.NonExplicitlyBoundClassError(provide_loc, cls)
        try:
            return self._obj_provider.provide_class(
                cls, self._injection_context_factory.new(cls.__init__),
                direct_init_pargs=[], direct_init_kwargs={})
        except errors.Error as e:
            if self._use_short_stack_traces:
                raise e
            else:
                raise



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from . import errors
from . import locations
from . import scoping


class InjectionContextFactory(object):
    """A creator of _InjectionContexts."""

    def __init__(self, is_scope_usable_from_scope_fn):
        """Initializer.

        Args:
          is_scope_usable_from_scope_fn: a function taking two scope IDs and
              returning whether an object in the first scope can be injected
              into an object from the second scope
        """
        self._is_scope_usable_from_scope_fn = is_scope_usable_from_scope_fn

    def new(self, injection_site_fn):
        """Creates a _InjectionContext.

        Args:
          injection_site_fn: the initial function being injected into
        Returns:
          a new empty _InjectionContext in the default scope
        """
        return _InjectionContext(
            injection_site_fn, binding_stack=[], scope_id=scoping.UNSCOPED,
            is_scope_usable_from_scope_fn=self._is_scope_usable_from_scope_fn)


class _InjectionContext(object):
    """The context of dependency-injecting some bound value."""

    def __init__(self, injection_site_fn, binding_stack, scope_id,
                 is_scope_usable_from_scope_fn):
        """Initializer.

        Args:
          injection_site_fn: the function currently being injected into
          binding_stack: a sequence of the bindings whose use in injection is
              in-progress, from the highest level (first) to the current level
              (last)
          scope_id: the scope ID of the current (last) binding's scope
          is_scope_usable_from_scope_fn: a function taking two scope IDs and
              returning whether an object in the first scope can be injected
              into an object from the second scope
        """
        self._injection_site_fn = injection_site_fn
        self._binding_stack = binding_stack
        self._scope_id = scope_id
        self._is_scope_usable_from_scope_fn = is_scope_usable_from_scope_fn

    def get_child(self, injection_site_fn, binding):
        """Creates a child injection context.

        A "child" injection context is a context for a binding used to
        inject something into the current binding's provided value.

        Args:
          injection_site_fn: the child function being injected into
          binding: a Binding
        Returns:
          a new _InjectionContext
        """
        child_scope_id = binding.scope_id
        new_binding_stack = self._binding_stack + [binding]
        if binding in self._binding_stack:
            raise errors.CyclicInjectionError(new_binding_stack)
        if not self._is_scope_usable_from_scope_fn(
                child_scope_id, self._scope_id):
            raise errors.BadDependencyScopeError(
                self.get_injection_site_desc(),
                self._scope_id, child_scope_id, binding.binding_key)
        return _InjectionContext(
            injection_site_fn, new_binding_stack, child_scope_id,
            self._is_scope_usable_from_scope_fn)

    def get_injection_site_desc(self):
        """Returns a description of the current injection site."""
        return locations.get_name_and_loc(self._injection_site_fn)



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import inspect

LOCALS_TOKEN = '<locals>'


def get_loc(thing):
    try:
        return '{0}:{1}'.format(
            inspect.getfile(thing), inspect.getsourcelines(thing)[1])
    except (TypeError, IOError):
        return 'unknown location'


def get_name_and_loc(thing):
    try:
        type_name = _get_type_name(thing)
        class_name = '{0}.{1}'.format(type_name, thing.__name__)
    except (TypeError, IOError):
        class_name = '{0}.{1}'.format(
            inspect.getmodule(thing).__name__, thing.__name__)
    try:
        return '{0} at {1}:{2}'.format(class_name, inspect.getfile(thing),
                                       inspect.getsourcelines(thing)[1])
    except (TypeError, IOError):
        return class_name


def get_back_frame_loc():
    back_frame = inspect.currentframe().f_back.f_back
    return '{0}:{1}'.format(back_frame.f_code.co_filename,
                            back_frame.f_lineno)


def _get_type_name(target_thing):
    """
    Functions, bound methods and unbound methods change significantly in Python 3.

    For instance:

    class SomeObject(object):
        def method():
            pass

    In Python 2:
    - Unbound method inspect.ismethod(SomeObject.method) returns True
    - Unbound inspect.isfunction(SomeObject.method) returns False
    - Unbound hasattr(SomeObject.method, 'im_class') returns True
    - Bound method inspect.ismethod(SomeObject().method) returns True
    - Bound method inspect.isfunction(SomeObject().method) returns False
    - Bound hasattr(SomeObject().method, 'im_class') returns True

    In Python 3:
    - Unbound method inspect.ismethod(SomeObject.method) returns False
    - Unbound inspect.isfunction(SomeObject.method) returns True
    - Unbound hasattr(SomeObject.method, 'im_class') returns False
    - Bound method inspect.ismethod(SomeObject().method) returns True
    - Bound method inspect.isfunction(SomeObject().method) returns False
    - Bound hasattr(SomeObject().method, 'im_class') returns False

    This method tries to consolidate the approach for retrieving the
    enclosing type of a bound/unbound method and functions.
    """
    thing = target_thing
    if hasattr(thing, 'im_class'):
        # only works in Python 2
        return thing.im_class.__name__
    if inspect.ismethod(thing):
        for cls in inspect.getmro(thing.__self__.__class__):
            if cls.__dict__.get(thing.__name__) is thing:
                return cls.__name__
        thing = thing.__func__
    if inspect.isfunction(thing) and hasattr(thing, '__qualname__'):
        qualifier = thing.__qualname__
        if LOCALS_TOKEN in qualifier:
            return _get_local_type_name(thing)
        return _get_external_type_name(thing)
    return inspect.getmodule(target_thing).__name__


def _get_local_type_name(thing):
    qualifier = thing.__qualname__
    parts = qualifier.split(LOCALS_TOKEN, 1)
    type_name = parts[1].split('.')[1]
    if thing.__name__ == type_name:
        return inspect.getmodule(thing).__name__
    return type_name


def _get_external_type_name(thing):
    qualifier = thing.__qualname__
    name = qualifier.rsplit('.', 1)[0]
    if hasattr(inspect.getmodule(thing), name):
        cls = getattr(inspect.getmodule(thing), name)
        if isinstance(cls, type):
            return cls.__name__
    return inspect.getmodule(thing).__name__



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from . import support
from . import arg_binding_keys
from . import decorators
from . import errors


class ObjectProvider(object):

    def __init__(self, binding_mapping, bindable_scopes, allow_injecting_none):
        self._binding_mapping = binding_mapping
        self._bindable_scopes = bindable_scopes
        self._allow_injecting_none = allow_injecting_none

    def provide_from_arg_binding_key(
            self, injection_site_fn, arg_binding_key, injection_context):
        binding_key = arg_binding_key.binding_key
        binding = self._binding_mapping.get(
            binding_key, injection_context.get_injection_site_desc())
        scope = self._bindable_scopes.get_sub_scope(binding)
        def Provide(*pargs, **kwargs):
            # TODO(kurts): probably capture back frame's file:line for
            # DirectlyPassingInjectedArgsError.
            child_injection_context = injection_context.get_child(
                injection_site_fn, binding)
            provided = scope.provide(
                binding_key,
                lambda: binding.proviser_fn(child_injection_context, self,
                                            pargs, kwargs))
            if (provided is None) and not self._allow_injecting_none:
                raise errors.InjectingNoneDisallowedError(
                    binding.get_binding_target_desc_fn())
            return provided
        provider_indirection = arg_binding_key.provider_indirection
        try:
            provided = provider_indirection.StripIndirectionIfNeeded(Provide)
        except TypeError:
            # TODO(kurts): it feels like there may be other TypeErrors that
            # occur.  Instead, decorators.get_injectable_arg_binding_keys()
            # should probably do all appropriate validation?
            raise errors.OnlyInstantiableViaProviderFunctionError(
                injection_site_fn, arg_binding_key,
                binding.get_binding_target_desc_fn())
        return provided

    def provide_class(self, cls, injection_context,
                      direct_init_pargs, direct_init_kwargs):
        if support.is_constructor_defined(cls):
            init_pargs, init_kwargs = self.get_injection_pargs_kwargs(
                cls.__init__, injection_context,
                direct_init_pargs, direct_init_kwargs)
        else:
            init_pargs = direct_init_pargs
            init_kwargs = direct_init_kwargs
        return cls(*init_pargs, **init_kwargs)

    def call_with_injection(self, provider_fn, injection_context,
                            direct_pargs, direct_kwargs):
        pargs, kwargs = self.get_injection_pargs_kwargs(
            provider_fn, injection_context, direct_pargs, direct_kwargs)
        return provider_fn(*pargs, **kwargs)

    def get_injection_pargs_kwargs(self, fn, injection_context,
                                   direct_pargs, direct_kwargs):
        di_kwargs = arg_binding_keys.create_kwargs(
            decorators.get_injectable_arg_binding_keys(
                fn, direct_pargs, direct_kwargs),
            lambda abk: self.provide_from_arg_binding_key(
                fn, abk, injection_context))
        duplicated_args = set(di_kwargs.keys()) & set(direct_kwargs.keys())
        if duplicated_args:
            raise errors.DirectlyPassingInjectedArgsError(
                duplicated_args, injection_context.get_injection_site_desc(),
                fn)
        all_kwargs = dict(di_kwargs)
        all_kwargs.update(direct_kwargs)
        return direct_pargs, all_kwargs



#!/usr/bin/env python
#
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from setuptools import setup
import semver


def versioning(version: str) -> str:
    """
    version to specification
    Author: Huan <zixia@zixia.net> (https://github.com/huan)

    X.Y.Z -> X.Y.devZ

    """
    sem_ver = semver.parse(version)

    major = sem_ver['major']
    minor = sem_ver['minor']
    patch = str(sem_ver['patch'])

    if minor % 2:
        patch = 'dev' + patch

    fin_ver = '%d.%d.%s' % (
        major,
        minor,
        patch,
    )

    return fin_ver


# Extra __version__ from VERSION instead of pinject.version so pip doen't
# import __init__.py and bump into dependencies that haven't been installed
# yet.
def get_version() -> str:
    """
    read version from VERSION file
    """
    version = '0.0.0'

    with open(
            os.path.join(
                os.path.dirname(__file__),
                'VERSION'
            )
    ) as handle:
        # Get X.Y.Z
        version = handle.read().strip()
        # versioning from X.Y.Z to X.Y.devZ
        version = versioning(version)

    return version


setup(
    name='pinject',
    version=get_version(),
    description='A pythonic dependency injection library',
    author='Kurt Steinkraus',
    author_email='kurt@steinkraus.us',
    url='https://github.com/google/pinject',
    license='Apache License 2.0',
    long_description=open('README.rst').read(),
    platforms='all',
    packages=['pinject'],
    install_requires=['six>=1.7.3', 'decorator>=4.3.0'],
)



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


from . import binding_keys
from . import provider_indirections


class ArgBindingKey(object):
    """The binding key for an arg of a function."""

    def __init__(self, arg_name, binding_key, provider_indirection):
        self._arg_name = arg_name
        self.binding_key = binding_key
        self.provider_indirection = provider_indirection

    def __repr__(self):
        return '<{0}>'.format(self)

    def __str__(self):
        return 'the arg named "{0}" {1}'.format(
            self._arg_name, self.binding_key.annotation_as_adjective())

    def __eq__(self, other):
        return (isinstance(other, ArgBindingKey) and
                self._arg_name == other._arg_name and
                self.binding_key == other.binding_key and
                self.provider_indirection == other.provider_indirection)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        # Watch out: self._arg_name is likely also binding_key._name, and so
        # XORing their hashes will remove the arg name from the hash.
        # self._arg_name is really captured as part of the other two, so let's
        # omit it.
        return (hash(self.binding_key) ^
                hash(self.provider_indirection))

    # TODO(kurts): the methods feel unbalanced: they only use self._arg_name.
    # That should probably be a full-fledged class, and ArgBindingKey should
    # just have two public attributes?

    def can_apply_to_one_of_arg_names(self, arg_names):
        """Returns whether this object can apply to one of the arg names."""
        return self._arg_name in arg_names

    def conflicts_with_any_arg_binding_key(self, arg_binding_keys):
        """Returns whether this arg binding key conflicts with others.

        One arg binding key conflicts with another if they are for the same
        arg name, regardless of whether they have the same annotation (or lack
        thereof).

        Args:
          arg_binding_keys: a sequence of ArgBindingKey
        Returns:
          True iff some element of arg_binding_keys is for the same arg name
              as this binding key
        """
        return self._arg_name in [abk._arg_name for abk in arg_binding_keys]


# TODO(kurts): Get a second opinion on module-level methods operating on
# internal state of classes.  In another language, this would be a static
# member and so allowed access to internals.
def get_unbound_arg_names(arg_names, arg_binding_keys):
    """Determines which args have no arg binding keys.

    Args:
      arg_names: a sequence of the names of possibly bound args
      arg_binding_keys: a sequence of ArgBindingKey each of whose arg names is
          in arg_names
    Returns:
      a sequence of arg names that is a (possibly empty, possibly non-proper)
          subset of arg_names

    """
    bound_arg_names = [abk._arg_name for abk in arg_binding_keys]
    return [arg_name for arg_name in arg_names
            if arg_name not in bound_arg_names]


def create_kwargs(arg_binding_keys, provider_fn):
    """Creates a kwargs map for the given arg binding keys.

    Args:
      arg_binding_keys: a sequence of ArgBindingKey for some function's args
      provider_fn: a function that takes an ArgBindingKey and returns whatever
          is bound to that binding key
    Returns:
      a (possibly empty) map from arg name to provided value
    """
    return {arg_binding_key._arg_name: provider_fn(arg_binding_key)
            for arg_binding_key in arg_binding_keys}


_PROVIDE_PREFIX = 'provide_'
_PROVIDE_PREFIX_LEN = len(_PROVIDE_PREFIX)


def new(arg_name, annotated_with=None):
    """Creates an ArgBindingKey.

    Args:
      arg_name: the name of the bound arg
      annotation: an Annotation, or None to create an unannotated arg binding
          key
    Returns:
      a new ArgBindingKey
    """
    if arg_name.startswith(_PROVIDE_PREFIX):
        binding_key_name = arg_name[_PROVIDE_PREFIX_LEN:]
        provider_indirection = provider_indirections.INDIRECTION
    else:
        binding_key_name = arg_name
        provider_indirection = provider_indirections.NO_INDIRECTION
    binding_key = binding_keys.new(binding_key_name, annotated_with)
    return ArgBindingKey(arg_name, binding_key, provider_indirection)



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import inspect
import sys


ALL_IMPORTED_MODULES = object()


def find_classes(modules, classes):
    if classes is not None:
        all_classes = set(classes)
    else:
        all_classes = set()
    for module in _get_explicit_or_default_modules(modules):
        # TODO(kurts): how is a module getting to be None??
        if module is not None:
            all_classes |= _find_classes_in_module(module)
    return all_classes


def _get_explicit_or_default_modules(modules):
    if modules is ALL_IMPORTED_MODULES:
        return list(sys.modules.values())
    elif modules is None:
        return []
    return modules


def _find_classes_in_module(module):
    classes = set()
    for member_name, member in inspect.getmembers(module):
        try:
            if inspect.isclass(member) and not member_name == '__class__':
                classes.add(member)
        except NameError:
            # In Python 3 calling isinstance() on SWIG's global cvar property
            # raises:
            #     "NameError: Unknown C global variable"
            # In that case just continue, otherwise let the Error through.
            if not member_name == 'cvar':
                raise
    return classes



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import decorator

from . import arg_binding_keys
from . import support
from . import errors
from . import locations
from . import scoping

_ARG_BINDING_KEYS_ATTR = '_pinject_arg_binding_keys'
_IS_WRAPPER_ATTR = '_pinject_is_wrapper'
_NON_INJECTABLE_ARG_NAMES_ATTR = '_pinject_non_injectables'
_ORIG_FN_ATTR = '_pinject_orig_fn'
_PROVIDER_DECORATIONS_ATTR = '_pinject_provider_decorations'


def annotate_arg(arg_name, with_annotation):
    """Adds an annotation to an injected arg.

    arg_name must be one of the named args of the decorated function, i.e.,
      @annotate_arg('foo', with_annotation='something')
      def a_function(foo):  # ...
    is OK, but
      @annotate_arg('foo', with_annotation='something')
      def a_function(bar, **kwargs):  # ...
    is not.

    The same arg (on the same function) may not be annotated twice.

    Args:
      arg_name: the name of the arg to annotate on the decorated function
      with_annotation: an annotation object
    Returns:
      a function that will decorate functions passed to it
    """
    arg_binding_key = arg_binding_keys.new(arg_name, with_annotation)
    return _get_pinject_wrapper(locations.get_back_frame_loc(),
                                arg_binding_key=arg_binding_key)


def inject(arg_names=None, all_except=None):
    """Marks an initializer explicitly as injectable.

    An initializer marked with @inject will be usable even when setting
    only_use_explicit_bindings=True when calling new_object_graph().

    This decorator can be used on an initializer or provider method to
    separate the injectable args from the args that will be passed directly.
    If arg_names is specified, then it must be a sequence, and only those args
    are injected (and the rest must be passed directly).  If all_except is
    specified, then it must be a sequence, and only those args are passed
    directly (and the rest must be specified).  If neither arg_names nor
    all_except are specified, then all args are injected (and none may be
    passed directly).

    arg_names or all_except, when specified, must not be empty and must
    contain a (possibly empty, possibly non-proper) subset of the named args
    of the decorated function.  all_except may not be all args of the
    decorated function (because then why call that provider method or
    initialzer via Pinject?).  At most one of arg_names and all_except may be
    specified.  A function may be decorated by @inject at most once.

    """
    back_frame_loc = locations.get_back_frame_loc()
    if arg_names is not None and all_except is not None:
        raise errors.TooManyArgsToInjectDecoratorError(back_frame_loc)
    for arg, arg_value in [('arg_names', arg_names), ('all_except', all_except)]:
        if arg_value is not None:
            if not arg_value:
                raise errors.EmptySequenceArgError(back_frame_loc, arg)
            if (not support.is_sequence(arg_value) or
                    support.is_string(arg_value)):
                raise errors.WrongArgTypeError(
                    arg, 'sequence (of arg names)', type(arg_value).__name__)
    if arg_names is None and all_except is None:
        all_except = []
    return _get_pinject_wrapper(
        back_frame_loc, inject_arg_names=arg_names,
        inject_all_except_arg_names=all_except)


def injectable(fn):
    """Deprecated.  Use @inject() instead.

    TODO(kurts): remove after 2014/6/30.
    """
    return inject()(fn)


def provides(arg_name=None, annotated_with=None, in_scope=None):
    """Modifies the binding of a provider method.

    If arg_name is specified, then the created binding is for that arg name
    instead of the one gotten from the provider method name (e.g., 'foo' from
    'provide_foo').

    If annotated_with is specified, then the created binding includes that
    annotation object.

    If in_scope is specified, then the created binding is in the scope with
    that scope ID.

    At least one of the args must be specified.  A provider method may not be
    decorated with @provides() twice.

    Args:
      arg_name: the name of the arg to annotate on the decorated function
      annotated_with: an annotation object
      in_scope: a scope ID
    Returns:
      a function that will decorate functions passed to it
    """
    if arg_name is None and annotated_with is None and in_scope is None:
        raise errors.EmptyProvidesDecoratorError(locations.get_back_frame_loc())
    return _get_pinject_wrapper(locations.get_back_frame_loc(),
                                provider_arg_name=arg_name,
                                provider_annotated_with=annotated_with,
                                provider_in_scope_id=in_scope)


class ProviderDecoration(object):
    """The provider method-relevant info set by @provides.

    Attributes:
      arg_name: the name of the arg provided by the provider function
      annotated_with: an annotation object
      in_scope_id: a scope ID
    """

    def __init__(self, arg_name, annotated_with, in_scope_id):
        self.arg_name = arg_name
        self.annotated_with = annotated_with
        self.in_scope_id = in_scope_id

    def __eq__(self, other):
        return (self.arg_name == other.arg_name and
                self.annotated_with == other.annotated_with and
                self.in_scope_id == other.in_scope_id)

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return (hash(self.arg_name) ^ hash(self.annotated_with) ^
                hash(self.in_scope_id))


def get_provider_fn_decorations(provider_fn, default_arg_names):
    """Retrieves the provider method-relevant info set by decorators.

    If any info wasn't set by decorators, then defaults are returned.

    Args:
      provider_fn: a (possibly decorated) provider function
      default_arg_names: the (possibly empty) arg names to use if none were
          specified via @provides()
    Returns:
      a sequence of ProviderDecoration
    """
    if hasattr(provider_fn, _IS_WRAPPER_ATTR):
        provider_decorations = getattr(provider_fn, _PROVIDER_DECORATIONS_ATTR)
        if provider_decorations:
            expanded_provider_decorations = []
            for provider_decoration in provider_decorations:
                # TODO(kurts): seems like default scope should be done at
                # ProviderDecoration instantiation time.
                if provider_decoration.in_scope_id is None:
                    provider_decoration.in_scope_id = scoping.DEFAULT_SCOPE
                if provider_decoration.arg_name is not None:
                    expanded_provider_decorations.append(provider_decoration)
                else:
                    expanded_provider_decorations.extend(
                        [ProviderDecoration(default_arg_name,
                                            provider_decoration.annotated_with,
                                            provider_decoration.in_scope_id)
                         for default_arg_name in default_arg_names])
            return expanded_provider_decorations
    return [ProviderDecoration(default_arg_name,
                               annotated_with=None,
                               in_scope_id=scoping.DEFAULT_SCOPE)
            for default_arg_name in default_arg_names]


def _get_pinject_decorated_fn(fn):
    if hasattr(fn, _IS_WRAPPER_ATTR):
        pinject_decorated_fn = fn
    else:
        def _pinject_decorated_fn(fn_to_wrap, *pargs, **kwargs):
            return fn_to_wrap(*pargs, **kwargs)

        pinject_decorated_fn = decorator.decorator(_pinject_decorated_fn, fn)
        # TODO(kurts): split this so that __init__() decorators don't get
        # the provider attribute.
        setattr(pinject_decorated_fn, _ARG_BINDING_KEYS_ATTR, [])
        setattr(pinject_decorated_fn, _IS_WRAPPER_ATTR, True)
        setattr(pinject_decorated_fn, _ORIG_FN_ATTR, fn)
        setattr(pinject_decorated_fn, _PROVIDER_DECORATIONS_ATTR, [])
    return pinject_decorated_fn


# TODO(kurts): separate out the parts for different decorators.
def _get_pinject_wrapper(
        decorator_loc, arg_binding_key=None, provider_arg_name=None,
        provider_annotated_with=None, provider_in_scope_id=None,
        inject_arg_names=None, inject_all_except_arg_names=None):
    def get_pinject_decorated_fn_with_additions(fn):
        pinject_decorated_fn = _get_pinject_decorated_fn(fn)
        orig_arg_names, unused_varargs, unused_keywords, unused_defaults = (
            support.get_method_args(getattr(pinject_decorated_fn, _ORIG_FN_ATTR)))
        if arg_binding_key is not None:
            if not arg_binding_key.can_apply_to_one_of_arg_names(orig_arg_names):
                raise errors.NoSuchArgToInjectError(decorator_loc, arg_binding_key, fn)
            if arg_binding_key.conflicts_with_any_arg_binding_key(
                getattr(pinject_decorated_fn, _ARG_BINDING_KEYS_ATTR)):
                raise errors.MultipleAnnotationsForSameArgError(
                    arg_binding_key, decorator_loc)
            getattr(pinject_decorated_fn, _ARG_BINDING_KEYS_ATTR).append(
                arg_binding_key)
        if (provider_arg_name is not None or
            provider_annotated_with is not None or
                provider_in_scope_id is not None):
            provider_decorations = getattr(
                pinject_decorated_fn, _PROVIDER_DECORATIONS_ATTR)
            provider_decorations.append(ProviderDecoration(
                provider_arg_name, provider_annotated_with,
                provider_in_scope_id))
        if (inject_arg_names is not None or
                inject_all_except_arg_names is not None):
            if hasattr(pinject_decorated_fn, _NON_INJECTABLE_ARG_NAMES_ATTR):
                raise errors.DuplicateDecoratorError('inject', decorator_loc)
            non_injectable_arg_names = []
            setattr(pinject_decorated_fn, _NON_INJECTABLE_ARG_NAMES_ATTR,
                    non_injectable_arg_names)
            if inject_arg_names is not None:
                non_injectable_arg_names[:] = [
                    x for x in orig_arg_names if x not in inject_arg_names]
                arg_names_to_verify = inject_arg_names
            else:
                non_injectable_arg_names[:] = inject_all_except_arg_names
                arg_names_to_verify = inject_all_except_arg_names
            for arg_name in arg_names_to_verify:
                if arg_name not in orig_arg_names:
                    raise errors.NoSuchArgError(decorator_loc, arg_name)
            if len(non_injectable_arg_names) == len(orig_arg_names):
                raise errors.NoRemainingArgsToInjectError(decorator_loc)
        return pinject_decorated_fn

    return get_pinject_decorated_fn_with_additions


def is_explicitly_injectable(cls):
    return (hasattr(cls, '__init__') and
            hasattr(cls.__init__, _IS_WRAPPER_ATTR))


def get_injectable_arg_binding_keys(fn, direct_pargs, direct_kwargs):
    non_injectable_arg_names = []
    if hasattr(fn, _IS_WRAPPER_ATTR):
        existing_arg_binding_keys = getattr(fn, _ARG_BINDING_KEYS_ATTR)
        orig_fn = getattr(fn, _ORIG_FN_ATTR)
        if hasattr(fn, _NON_INJECTABLE_ARG_NAMES_ATTR):
            non_injectable_arg_names = getattr(
                fn, _NON_INJECTABLE_ARG_NAMES_ATTR)
    else:
        existing_arg_binding_keys = []
        orig_fn = fn

    arg_names, unused_varargs, unused_keywords, defaults = support.get_method_args(orig_fn)
    num_args_with_defaults = len(defaults) if defaults is not None else 0
    if num_args_with_defaults:
        arg_names = arg_names[:-num_args_with_defaults]
    unbound_injectable_arg_names = arg_binding_keys.get_unbound_arg_names(
        [arg_name for arg_name in _remove_self_if_exists(arg_names)
         if arg_name not in non_injectable_arg_names],
        existing_arg_binding_keys)

    all_arg_binding_keys = list(existing_arg_binding_keys)
    all_arg_binding_keys.extend([arg_binding_keys.new(arg_name)
                                 for arg_name in unbound_injectable_arg_names])
    return all_arg_binding_keys


# TODO(kurts): this feels icky.  Is there no way around this, because
# cls.__init__() takes self but instance.__init__() doesn't, and python is
# awkward here?
def _remove_self_if_exists(args):
    if args and args[0] == 'self':
        return args[1:]
    else:
        return args



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import threading

from . import errors


class _SingletonScopeId(object):
    def __str__(self):
        return 'singleton scope'
SINGLETON = _SingletonScopeId()


class _PrototypeScopeId(object):
    def __str__(self):
        return 'prototype scope'
PROTOTYPE = _PrototypeScopeId()


DEFAULT_SCOPE = SINGLETON
_BUILTIN_SCOPES = [SINGLETON, PROTOTYPE]


class Scope(object):

    def provide(self, binding_key, default_provider_fn):
        raise NotImplementedError()


class PrototypeScope(object):

    def provide(self, binding_key, default_provider_fn):
        return default_provider_fn()


class SingletonScope(object):

    def __init__(self):
        self._binding_key_to_instance = {}
        self._providing_binding_keys = set()
        # The lock is re-entrant so that default_provider_fn can provide
        # something else in singleton scope.
        self._rlock = threading.RLock()

    def provide(self, binding_key, default_provider_fn):
        with self._rlock:
            try:
                return self._binding_key_to_instance[binding_key]
            except KeyError:
                instance = default_provider_fn()
                self._binding_key_to_instance[binding_key] = instance
                return instance


class _UnscopedScopeId(object):
    def __str__(self):
        return 'unscoped scope'
UNSCOPED = _UnscopedScopeId()


def get_id_to_scope_with_defaults(id_to_scope=None):
    if id_to_scope is not None:
        for scope_id in _BUILTIN_SCOPES:
            if scope_id in id_to_scope:
                raise errors.OverridingDefaultScopeError(scope_id)
        id_to_scope = dict(id_to_scope)
    else:
        id_to_scope = {}
    id_to_scope[PROTOTYPE] = PrototypeScope()
    id_to_scope[SINGLETON] = SingletonScope()
    return id_to_scope


# TODO(kurts): either make this class pull its weight, or delete it.
class BindableScopes(object):

    def __init__(self, id_to_scope):
        self._id_to_scope = id_to_scope

    def get_sub_scope(self, binding):
        return self._id_to_scope[binding.scope_id]



"""Copyright 2013 Google Inc. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""


import sys
from .version import (
    VERSION,
)

__version__ = VERSION

__all__ = [
    '__version__',
]

from .bindings import BindingSpec
__all__.extend(['BindingSpec'])
from .decorators import annotate_arg, inject, injectable, provides
__all__.extend(['annotate_arg', 'inject', 'injectable', 'provides'])
for thing_name in dir(errors):
    thing = getattr(errors, thing_name)
    if type(thing) == type(str):
        setattr(sys.modules[__name__], thing_name, thing)
        __all__.append(thing_name)
from .initializers import copy_args_to_internal_fields
from .initializers import copy_args_to_public_fields
from .object_graph import new_object_graph
__all__.extend(['new_object_graph'])
from .scoping import PROTOTYPE, Scope, SINGLETON
__all__.extend(['PROTOTYPE', 'Scope', 'SINGLETON'])

# TODO(kurts): figure out how to avoid breaking unittests by uncommenting this
#   section.
# this_module = sys.modules[__name__]
# for thing_name in dir(this_module):
#     if (thing_name != 'this_module' and
#         type(getattr(this_module, thing_name)) == type(this_module)):
#         delattr(this_module, thing_name)
# delattr(this_module, '__builtins__')
# delattr(this_module, 'thing_name')
# delattr(this_module, 'this_module')
